module ActionController
  module Macros
    module ImageMagick #:nodoc:
      def self.append_features(base) #:nodoc:
        super
        
        base.extend(ClassMethods)
      end

      # This macro can be used to add ImageMagick functionality to a controller.
      # You can use it to access ImageMagick commands with a simple template tag.
      # ImageMagick is a set of image processing tools, that can do nearly everything
      # you'd ever want to do with an image. Resize, rotate, crop and much more.
      #
      # This needs the RMagick library (and ImageMagick, of course) to work.
      #
      # * http://www.imagemagick.org/
      # * http://rmagick.rubyforge.org/
      #
      # Example:
      #
      #   # Controller
      #   class Photo < ApplicationController
      #     imagemagick_for '/var/lib/photos'
      #   end
      #
      #   # View
      #   <%= imagemagick_tag 'presentation.jpg', 'resize(100x50)+flop' %>
      #
      # This example generates an <tt>&lt;img&gt;</tt>-tag in the view, which points to
      # the +imagemagick+ action of the Photo controller. All you need to do to
      # generate this controller, is adding the <tt>imagemagick_for</tt> method to your
      # controller class. The +imagemagick+ action opens the image, sends the
      # commands to RMagick for processing and then sends the result to the browser.
      #
      # This example shows you the image presentation.jpg from the /var/lib/photos
      # directory. It is processed with the commands +resize(100x50)+, which means
      # that ImageMagick will shrink it to fit it in an area of 100x50 pixels, and
      # +flop+, which means that left and right are reversed.
      #
      #
      # == The image directory
      #
      # This extension expects to find all images in one directory, the name of which
      # is specified with the +imagemagick_for+ method. In this directory, you can simply
      # place the image files. The names and extensions are not really important,
      # ImageMagick is smart enough to determine the file type from the contents of the file.
      #
      # You can use the ImageMagick extension in multiple controllers, but you might
      # want to specify a different image directory for each of them. (But that's not
      # compulsory.)
      #
      #
      # == Caching
      #
      # If you don't take any precautions, the rendered image will not be saved, and
      # it has to be rendered from scratch for every new request. This can make it a
      # little slow. This extension therefore provides a caching mechanism, which caches
      # the processed images. For the following requests, it simply sends the cached
      # image, instead of re-rendering it.
      #
      # To enable caching, you need a cache directory that is writable by the webserver.
      # If you have that directory, all you have to do is specify it with the +cache+ option
      # of the +imagemagick_for+ method.
      #
      #   # Controller
      #   class Photo < ApplicationController
      #     imagemagick_for '/var/lib/photos', :cache=>'/var/lib/photo-cache'
      #   end
      #
      # If you enable caching, you need a separate cache directory for every controller
      # that uses the +imagemagick_tag+.
      #
      #
      # == Pre-rendering
      #
      # By default, the commands are transferred in the URL of the image, and are executed
      # when the user requests the image. This is reasonably safe, but it does allow your
      # visitors to specify any ImageMagick command they want by simply changing the URL.
      # If you think this is a problem, you can enable the pre-rendering function.
      #
      # Pre-rendering means that the image is rendered earlier in the process: when the
      # +imagemagick_tag+ is displayed by the view. The image is rendered and cached, and the
      # name of that cached image is sent to the browser, which then uses it to request the
      # image. This means that the user does not have the possibility to influence the
      # rendering of the image, as rendering is done from the view. It also means that the
      # initial delay of the rendering of the image is on the web page, instead of on the
      # loading of the images.
      #
      # The pre-rendered images are stored by the caching system, so you'll need to
      # enable caching if you want to use pre-rendering.
      #
      #   # Controller
      #   class Photo < ApplicationController
      #     imagemagick_for '/var/lib/photos', :cache=>'/var/lib/photo-cache', :prerender=>true
      #   end
      # 
      #
      # == Routes.rb trick
      #
      # +imagemagick_for+ works without any additional routes. It uses urls like:
      #   /photo/imagemagick/presentation.jpg?command=resize(100x100)
      #
      # If you'd like to get rid of the <tt>?command=</tt>, you can add the following routes to
      # your <tt>routes.rb</tt>:
      #     map.connect ':controller/imagemagick/:id', :action=>'imagemagick', :commands=>''
      #     map.connect ':controller/imagemagick/:commands/:id', :action=>'imagemagick'
      #
      # The above url will now be rendered as:
      #   /photo/imagemagick/resize(100x100)/presentation.jpg
      #
      module ClassMethods
        #
        # Enables the ImageMagick extension for this controller. The only thing you
        # really need is the +image_path+ parameter, which points to the directory with
        # images for this controller. If you want to, you can also specify the following +options+:
        #
        # * <tt>:cache</tt>: the directory where cached images are stored, or +nil+ (the default)
        #   if you don't want caching. The directory has to be writable by the web server.
        # * <tt>:prerender</tt>: set this to +true+ if you want pre-rendering. This only works if
        #   you have enabled caching. By default, pre-rendering is off.
        #
        def imagemagick_for(image_path, options = {})
          # strip the trailing slash
          image_path = image_path.sub(/\/$/, '') if image_path =~ /\/$/
          
          require 'RMagick'
          
          configuration = { :image_path=>image_path, :cache => nil, :prerender=>false }
          configuration.update(options)
          
          #
          # Every controller-class gets a imagemagick_macro_helper-variable,
          # which contains the configuration options.
          #
          instance_variable_set(:@imagemagick_macro_helper, ImageMagickMacroHelper.new(configuration))
          
          #
          # Makes the class helper available from the controller instance.
          #
          define_method(:imagemagick_macro_helper) do
            return self.class.instance_variable_get(:@imagemagick_macro_helper)
          end
          protected :imagemagick_macro_helper
          
          #
          # Shortcut to the +url_for_imagemagick+ method of the helper.
          # Gives the url to the +imagemagick+ action for this filename, with these commands.
          #
          define_method(:url_for_imagemagick) do |filename, p_commands|
            url_for(imagemagick_macro_helper.url_options_for_imagemagick(filename, p_commands))
          end
          
          #
          # Where it all takes place: the +imagemagick+ action of the controller.
          # Reads the filename and the commands from the request and sends an image.
          #
          define_method(:imagemagick) do
            commands = ImageMagickMacroHelper::MagickCommandList.from_anything(@params["commands"])
            filename = File.basename(@params["id"])
            
            # is prerendering enabled?
            if imagemagick_macro_helper.prerender?
              # yes, we only accept requests for images that are already in the cache
              cache_path = "#{configuration[:cache]}/" + imagemagick_macro_helper.cache_filename(filename, commands)
              
              if File.exists?(cache_path)
                # determine the mime type and send the file
                @response.headers["Content-type"] = Magick::Image.read(cache_path).first.mime_type
                render :file=>cache_path
              else
                render :text=>"Image not found.", :status=>404
              end
              
            else
              # no, we might have to render the image
              pathname = "#{image_path}/#{filename}"
              
              if filename =~ /^\.*$/ || !File.readable?(pathname)
                render :text=>"Image not found.", :status=>404
              else
                # the helper finds and/or renders the image
                picture = imagemagick_macro_helper.render(filename, commands)
                
                if picture.is_a?(String)
                  # the image is already cached, render from file
                  @response.headers["Content-type"] = Magick::Image.read(picture).first.mime_type
                  render :file=>picture
                  
                else
                  # the image was not yet cached, render from memory
                  @response.headers["Content-type"] = picture.mime_type
                  render :text=>picture.to_blob
                end
              end
            end
          end
          
        end
      end
      
      #
      # The class that contains all methods of the ImageMagick extension.
      # Every controller with imagemagick_for gets an instance of this, filled
      # with its own configuration options.
      #
      class ImageMagickMacroHelper #:nodoc: all
        def initialize(configuration)
          @configuration = configuration
        end
        
        def caching?
          return !@configuration[:cache].nil?
        end
        
        def prerender?
          # only if caching is available
          return caching? && @configuration[:prerender]
        end
        
        #
        # Returns a Hash of url_for options that render the given filename
        # with the given ImageMagick commands.
        #
        def url_options_for_imagemagick(filename, p_commands)
          commands = MagickCommandList.from_anything(p_commands)
          path_options = { :action=>"imagemagick" }
          
          if prerender?
            path_options[:id] = cache_filename(filename, commands)
            render(filename, commands)
          else
            render(filename, commands)
            path_options[:id] = filename
            path_options[:commands] = commands.to_s unless commands.commands.empty?
          end
          
          return path_options
        end
        
        #
        # Returns the filename the cached image will have if it is renderd
        # with the given commands.
        #
        def cache_filename(filename, commands)
          commands = MagickCommandList.from_anything(commands)
          return commands.to_cache_s + filename
        end
        
        #
        # Renders the image with the given commands.
        # The result is saved in the cache, if that is enabled
        #
        # The function returns either an Magick::Image-object of the image or
        # a String, which points to the already-cached image.
        #
        def render(filename, commands)
          pathname = @configuration[:image_path] + "/"+ filename
          
          cache_path = nil
          if caching?
            cache_path = @configuration[:cache] + "/" + cache_filename(filename, commands)
          end 
          
          if (!caching? || !File.exists?(cache_path) || File.mtime(cache_path) < File.mtime(pathname))
            pic = Magick::Image.read(pathname).first
            
            commands.execute_on(pic)
            
            pic.strip!
            
            if caching?
              pic.write(cache_path)
            end
            
            return pic
          else
            return cache_path
          end
        end
        
        #
        # Contains a list of ImageMagick commands (of the type MagickCommand).
        #
        class MagickCommandList
          attr_reader :commands
          
          def initialize
            @commands = Array.new
          end
          
          #
          # Returns a String of the commands for use in a url.
          # (The commands are chained with a +.)
          #    => resize(100x1)+chop
          #
          def to_s
            command_string = @commands.join("+")
          end
          
          #
          # Returns a String of the commands for use in the cache-filename.
          # (All strange characters are removed.)
          #    => resize.100x1.chop.
          #
          def to_cache_s
            if !commands.empty?
              command_string = to_s + "."
              command_string.gsub!(/[+,]/, '.')
              command_string.gsub!(/[^a-zA-Z0-9.]/, '')
            else
              command_string = ""
            end
            return command_string
          end
          
          #
          # Executes the commands on this list on the picture.
          # picture should be an Magick::Image-object.
          #
          def execute_on(picture)
            @commands.each { |cmd| cmd.execute_on(picture) }
          end
          
          #
          # Parses a String of commands and returns the MagickCommandList.
          # String is of the type generated by MagickCommandList::to_s
          #     resize(100x1)+chop
          #
          def self.from_s(command_string)
            list = MagickCommandList.new
            (command_string+"+").scan(/(([^(+]+)(\([^)]+\))?)*/) do |match|
              list.commands << MagickCommand.from_s(match[0]) unless match[0].nil?
            end
            return list
          end
          
          #
          # Returns the MagickCommandList for p_commands.
          # p_commands can be:
          #
          # * a +MagickCommandList+: it will be returned
          # * an +Array+: a list of commands. Commands can be Strings and
          #      arrays: [command, args]:
          #         [ "resize(1x100)", "resize", ["resize", 100] ]
          # * a +String+: will be parsed by MagickCommandList.from_s
          # * anything else: an empty MagickCommandList will be returned
          #
          def self.from_anything(p_commands)
            if p_commands.is_a?(MagickCommandList)
              commands = p_commands
            elsif p_commands.is_a?(Array)
              commands = MagickCommandList.new
              p_commands.each { |cmd| commands.commands << MagickCommand.new(cmd[0].to_s, cmd[1]) }
            elsif p_commands.is_a?(String)
              commands = MagickCommandList.from_s(p_commands)
            else
              commands = MagickCommandList.new
            end
            
            return commands
          end
        end
        
        #
        # A command for in a MagickCommandList.
        # (A command is a command string, and possibly a data string.)
        #
        class MagickCommand
          attr_accessor :command, :data
          
          def initialize(command, data = nil)
            @command = command
            @data = data
          end
          
          #
          # Converts the command to a String version for use in the image url.
          #    "command"  or  "command(args)"
          #
          def to_s
            command_string = @command
            command_string += "(" + MagickCommand.encode(data) + ")" unless data.nil?
            return command_string
          end
          
          #
          # Executes this command on the picture.
          #
          def execute_on(picture)
            case @command
              when "s", "resize", "geometry"
                picture.change_geometry!(@data) { |cols, rows, img| img.resize!(cols, rows) }
                
              when "blur", "blur_image"
                args = [ @data.split(/,/)[0].to_f,
                         @data.split(/,/)[1].to_f ]
                picture.blur_image(args[0], args[1])
                
              when "crop"
                args = [ @data.split(/,/)[0].to_i,
                         @data.split(/,/)[1].to_i,
                         @data.split(/,/)[2].to_i,
                         @data.split(/,/)[3].to_i ]
                picture.crop!(args[0], args[1], args[2], args[3])
              
              when "colorize"
                args = [ @data.split(/,/)[0].to_f,
                         @data.split(/,/)[1].to_f,
                         @data.split(/,/)[2].to_f,
                         @data.split(/,/)[3] ]
                args[3] = "#" + args[3] unless args[3][0,1]=="#"
                picture.colorize(args[0], args[1], args[2], args[3])
              
              when "equalize"
                picture = picture.equalize
              
              when "flip"
                picture.flip!
                
              when "flop"
                picture.flop!
                
              when "implode"
                args = [ @data.split(/,/)[0].to_f ]
                picture = picture.implode(args[0])
                
              when "level"
                args = [ @data.split(/,/)[0].to_f,
                         @data.split(/,/)[1].to_f ]
                picture = picture.level(args[0], args[1])
                
              when "normalize"
                picture = picture.normalize
                
              when "oilpaint"
                args = [ @data.split(/,/)[0].to_f ]
                picture = picture.oil_paint(args[0])
                
              when "opaque"
                args = [ @data.split(/,/)[0],
                         @data.split(/,/)[1] ]
                args[0] = "#" + args[0] unless args[0][0,1]=="#"
                args[1] = "#" + args[1] unless args[1][0,1]=="#"
                picture = picture.opaque(args[0], args[1])
                
              when "posterize"
                args = [ @data.split(/,/)[0].to_i ]
                picture = picture.posterize(args[0])
                
              when "sharpen"
                args = [ @data.split(/,/)[0].to_f,
                         @data.split(/,/)[1].to_f ]
                picture = picture.sharpen(args[0], args[1])
                
              when "rotate"
                args = [ @data.split(/,/)[0].to_i ]
                picture = picture.rotate!(args[0])
                
              when "border"
                args = [ @data.split(/,/)[0].to_i,
                         @data.split(/,/)[1].to_i,
                         @data.split(/,/)[2] ]
                args[2] = "#" + args[2] unless args[2][0,1]=="#"
                picture = picture.border!(args[0], args[1], args[2])
                
              when "trim"
                picture = picture.trim!
                
              when "part"
                # This method is not available from imagemagick, but
                # it is useful: it will always give you an image of
                # exactly the required size. part(100x100) gives you
                # a 100x100 pixel crop from the center of the image.
                #
                # For example: the image is   ********
                #                             ********
                #                             ********
                # 
                # And you want this:          **
                #                             **
                # It will then first resize the image: ****
                #                                      ****
                # And then crops it:         *[**]*
                #                            *[**]*
                #
                # If the source image was this:
                #        **                    --
                #        **  it would crop:    **
                #        **                    **
                #        **                    --
                #
                # The command also scales up, if the image is too small
                # to cover the whole part.
                
                target_w = @data.split(/x/)[0].to_i
                target_h = @data.split(/x/)[1].to_i
                orig_w = picture.columns
                orig_h = picture.rows
                
                if (orig_w.to_f/orig_h) > (target_w.to_f/target_h)
                  # scale to correct height
                  picture.change_geometry!("x"+target_h.to_s) { |cols, rows, img| img.resize!(cols, rows) }
                  
                  # remove left and right
                  picture.crop!((picture.columns - target_w) / 2, 0, target_w, target_h)
                  
                else
                  # scale to correct width
                  picture.change_geometry!(target_w.to_s+"x") { |cols, rows, img| img.resize!(cols, rows) }
                  
                  # remove top and bottom
                  picture.crop!(0, (picture.height - target_h) / 2, target_w, target_h)
                end
                
            end
          end
          
          #
          # Reads a command string (of the MagickCommand::to_s format) and returns
          # a MagickCommand.
          #   resize(100x10)  becomes @command = "resize", @data = "100x10"
          #
          def self.from_s(command)
            return MagickCommand.new(command[/^[^(]+/], command[/\((.+)\)$/, 1])
          end
          
          #
          # Encodes the data string for use in the url command string.
          #
          def self.encode(data)
            # in the command string, ( and ) are used to separate
            # the argument from the command: <tt>resize(100x10)</tt>
            return data.to_s.tr("()", "[]")
          end
          
          #
          # Decodes the data string from the url command string.
          #
          def self.decode(data)
            return data.to_s.tr("[]", "()")
          end
        end
      end
    end
  end
end